#define G_LOG_DOMAIN "ap-skill"

#include "ap-skill.h"

#include <glib.h>
#include <glib-object.h>

typedef struct _APSkillPrivate
{
    gchar    *id;
    gchar    *name;
    gint      score;
    gchar    *modifier;
    gint      bonus;
    gboolean  proficient;
} APSkillPrivate;

struct _APSkill
{
    GObject parent_instance;
};

G_DEFINE_TYPE_WITH_PRIVATE (APSkill, ap_skill, G_TYPE_OBJECT)

enum
{
    PROP_0,
    PROP_ID,
    PROP_NAME,
    PROP_SCORE,
    PROP_BONUS,
    PROP_MODIFIER,
    PROP_PROFICIENT,
    N_PROPERTIES
};

static GParamSpec *obj_properties[N_PROPERTIES] = { NULL, };

static void
ap_skill_set_property (GObject      *object,
                       guint         property_id,
                       const GValue *value,
                       GParamSpec   *pspec)
{
    APSkill *self = AP_SKILL (object);
    APSkillPrivate *priv = ap_skill_get_instance_private (self);

    switch (property_id)
    {
        case PROP_ID:
            priv->id = g_value_dup_string (value);
            break;

        case PROP_NAME:
            priv->name = g_value_dup_string (value);
            break;

        case PROP_SCORE:
            priv->score = g_value_get_int (value);
            break;

        case PROP_BONUS:
            priv->bonus = g_value_get_int (value);
            break;

        case PROP_MODIFIER:
            priv->modifier = g_value_dup_string (value);
            break;

        case PROP_PROFICIENT:
            priv->proficient = g_value_get_boolean (value);
            break;

        default:
            G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
            break;
    }
}

static void
ap_skill_get_property (GObject      *object,
                       guint         property_id,
                       GValue       *value,
                       GParamSpec   *pspec)
{
    APSkill *self = AP_SKILL (object);
    APSkillPrivate *priv = ap_skill_get_instance_private (self);

    switch (property_id)
    {
        case PROP_ID:
            g_value_set_string (value, priv->id);
            break;

        case PROP_NAME:
            g_value_set_string (value, priv->name);
            break;

        case PROP_SCORE:
            g_value_set_int (value, priv->score);
            break;

        case PROP_BONUS:
            g_value_set_int (value, priv->bonus);
            break;

        case PROP_MODIFIER:
            g_value_set_string (value, priv->modifier);
            break;

        case PROP_PROFICIENT:
            g_value_set_boolean (value, priv->proficient);
            break;

        default:
            G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
            break;
    }
}

static void
ap_skill_finalize (GObject *object)
{
    G_OBJECT_CLASS (ap_skill_parent_class)->finalize (object);
}

static void
ap_skill_dispose (GObject *object)
{
    G_OBJECT_CLASS (ap_skill_parent_class)->finalize (object);
}

static void
ap_skill_init (APSkill *self)
{
}

static void
ap_skill_class_init (APSkillClass *klass)
{
    GObjectClass *object_class = G_OBJECT_CLASS (klass);

    object_class->set_property = ap_skill_set_property;
    object_class->get_property = ap_skill_get_property;
    object_class->finalize = ap_skill_finalize;
    object_class->dispose = ap_skill_dispose;

	/**
	 * APSkill:id:
	 *
	 * ID of the skill, normally the skill name but no special characters,
     * lowercase and spaces replaced with dashes or underscores.
     *
     * ## Example
     * `Example Skill` becomes `example-skill`
	 */
    obj_properties[PROP_ID] =
        g_param_spec_string ("id",
                             "id",
                             "ID of the skill.",
                             NULL, /* default value */
                             G_PARAM_READWRITE);

	/**
	 * APSkill:name:
	 *
	 * Name of the skill.
	 */
    obj_properties[PROP_NAME] =
        g_param_spec_string ("name",
                             "name",
                             "Name of the skill.",
                             NULL  /* default value */,
                             G_PARAM_READWRITE);

	/**
	 * APSkill:score:
	 *
	 * Score of the skill.
	 */
    obj_properties[PROP_SCORE] =
        g_param_spec_int ("score",
                          "score",
                          "The skill score level.",
                          0,
                          100,
                          0, /* default value */
                          G_PARAM_READWRITE);

	/**
	 * APSkill:bonus:
	 *
	 * Skill bonus of the ability.
	 */
    obj_properties[PROP_BONUS] =
        g_param_spec_int ("bonus",
                          "bonus",
                          "The skill bonus.",
                          0,
                          100,
                          0, /* default value */
                          G_PARAM_READWRITE);

	/**
	 * APSkill:modifier:
	 *
	 * Modifier of the skill.
	 */
    obj_properties[PROP_MODIFIER] =
        g_param_spec_string ("modifier",
                             "modifier",
                             "The ability id that modifies the skill.",
                             NULL, /* default value */
                             G_PARAM_READWRITE);

	/**
	 * APSkill:proficient:
	 *
	 * Skill proficiency of the ability.
	 */
    obj_properties[PROP_PROFICIENT] =
        g_param_spec_boolean ("proficient",
                              "proficient",
                              "Is the character proficient with this skill?",
                              FALSE, /* default value */
                              G_PARAM_READWRITE);

    g_object_class_install_properties (object_class,
                                       N_PROPERTIES,
                                       obj_properties);
}

/*
 * *****************************************************************************
 * API
 * *****************************************************************************
 */
/**
 * ap_skill_new:
 *
 * Creates a new #APSkill
 *
 * Returns: #APSkill
 */
APSkill *
ap_skill_new (void)
{
    return g_object_new (AP_TYPE_SKILL,
                         NULL);
}

/**
 * ap_skill_set_id:
 * @self: An instance of #APSkill
 * @id: (gchar *): Internal name for the @self instance
 *
 * Set the internal skill name of @self, used for querying and inter-connecting
 * the various Adventure Pack modules. The ID should be lowercase, contain no
 * special character and spaces.
 *
 * ## Examples
 * * `Acrobatics` becomes `acrobatics`
 * * `Animal Handling` becomes `animal-handling`
 */
void
ap_skill_set_id (APSkill *self,
                 gchar   *id)
{
    GValue new_id = G_VALUE_INIT;

    g_assert (AP_IS_SKILL (self));
    g_assert (id != NULL);
    g_assert (g_strcmp0 (id, "") != 0);

    g_value_init (&new_id, G_TYPE_STRING);
    g_value_set_string (&new_id, id);

    g_object_set_property (G_OBJECT (self), "id", &new_id);

    g_value_unset (&new_id);
}

/**
 * ap_skill_get_id:
 * @self: An instance of #APSkill
 *
 * Gets the internal name (id) of the @self instance.
 *
 * Returns: (gchar *) (nullable): The internal name of @self (#APSkill)
 */
gchar *
ap_skill_get_id (APSkill *self)
{
    GValue val = G_VALUE_INIT;
    gchar *retval;

    g_assert (AP_IS_SKILL (self));

    g_value_init (&val, G_TYPE_STRING);

    g_object_get_property (G_OBJECT (self), "id", &val);

    retval = g_value_dup_string (&val);

    g_value_unset (&val);

    return retval;
}

/**
 * ap_skill_set_name:
 * @self: An instance of #APSkill
 * @name: (gchar *): The skill name for the @self instance
 *
 * Sets the skill name for @self.
 *
 * ## Examples
 * * `Acrobatics`
 * * `Animal Handling`
 */
void
ap_skill_set_name (APSkill *self,
                   gchar   *name)
{
    GValue new_name = G_VALUE_INIT;

    g_assert (AP_IS_SKILL (self));
    g_assert (name != NULL);
    g_assert (g_strcmp0 (name, "") != 0);

    g_value_init (&new_name, G_TYPE_STRING);
    g_value_set_string (&new_name, name);

    g_object_set_property (G_OBJECT (self), "name", &new_name);

    g_value_unset (&new_name);
}

/**
 * ap_skill_get_name:
 * @self: An instance of #APSkill
 *
 * Gets the name of the @self instance.
 *
 * Returns: (gchar *) (nullable): The skill name of @self (#APSkill)
 */
gchar *
ap_skill_get_name (APSkill *self)
{
    GValue val = G_VALUE_INIT;
    gchar *retval;

    g_assert (AP_IS_SKILL (self));

    g_value_init (&val, G_TYPE_STRING);

    g_object_get_property (G_OBJECT (self), "name", &val);

    retval = g_value_dup_string (&val);

    g_value_unset (&val);

    return retval;
}

/**
 * ap_skill_set_score:
 * @self: An instance of #APSkill
 * @score: (gint): The skill score for the @self instance
 *
 * Sets the skill score for @self.
 */
void
ap_skill_set_score (APSkill *self,
                    gint     score)
{
    GValue new_score = G_VALUE_INIT;

    g_assert (AP_IS_SKILL (self));

    g_value_init (&new_score, G_TYPE_INT);
    g_value_set_int (&new_score, score);

    g_object_set_property (G_OBJECT (self), "score", &new_score);

    g_value_unset (&new_score);
}

/**
 * ap_skill_get_score:
 * @self: An instance of #APSkill
 *
 * Gets the ability score of the @self instance.
 *
 * Returns: (gint) (nullable): The ability score of @self (#APSkill)
 */
gint
ap_skill_get_score (APSkill *self)
{
    GValue val = G_VALUE_INIT;
    gint retval;

    g_assert (AP_IS_SKILL (self));

    g_value_init (&val, G_TYPE_INT);

    g_object_get_property (G_OBJECT (self), "score", &val);

    retval = g_value_get_int (&val);

    g_value_unset (&val);

    return retval;
}

/**
 * ap_skill_set_bonus:
 * @self: An instance of #APSkill
 * @bonus: (gint): The skill score bonus for the @self instance
 *
 * Sets the skill score bonus for @self.
 */
void
ap_skill_set_bonus (APSkill *self,
                    gint     bonus)
{
    GValue new_bonus = G_VALUE_INIT;

    g_assert (AP_IS_SKILL (self));

    g_value_init (&new_bonus, G_TYPE_INT);
    g_value_set_int (&new_bonus, bonus);

    g_object_set_property (G_OBJECT (self), "bonus", &new_bonus);

    g_value_unset (&new_bonus);
}

/**
 * ap_skill_get_bonus:
 * @self: An instance of #APSkill
 *
 * Gets the skill score bonus of the @self instance.
 *
 * Returns: (gint) (nullable): The skill score bonus of @self (#APSkill)
 */
gint
ap_skill_get_bonus (APSkill *self)
{
    GValue val = G_VALUE_INIT;
    gint retval;

    g_assert (AP_IS_SKILL (self));

    g_value_init (&val, G_TYPE_INT);

    g_object_get_property (G_OBJECT (self), "bonus", &val);

    retval = g_value_get_int (&val);

    g_value_unset (&val);

    return retval;
}

/**
 * ap_skill_set_modifier:
 * @self: An instance of #APSkill
 * @bonus: (gchar *): The ability id that modifies for the @self instance
 *
 * Sets the ability id that modifies for @self.
 */
void
ap_skill_set_modifier (APSkill *self,
                       gchar   *modifier)
{
    GValue new_modifier = G_VALUE_INIT;

    g_assert (AP_IS_SKILL (self));

    g_value_init (&new_modifier, G_TYPE_STRING);
    g_value_set_string (&new_modifier, modifier);

    g_object_set_property (G_OBJECT (self), "modifier", &new_modifier);

    g_value_unset (&new_modifier);
}

/**
 * ap_ability_get_modifier:
 * @self: An instance of #APSkill
 *
 * Gets the ability modifier id of the @self instance.
 *
 * Returns: (gchar *) (nullable): The ability modifier id of @self (#APSkill)
 */
gchar *
ap_skill_get_modifier (APSkill *self)
{
    GValue val = G_VALUE_INIT;
    gchar *retval;

    g_assert (AP_IS_SKILL (self));

    g_value_init (&val, G_TYPE_STRING);

    g_object_get_property (G_OBJECT (self), "modifier", &val);

    retval = g_value_dup_string (&val);

    g_value_unset (&val);

    return retval;
}

/**
 * ap_skill_set_bonus:
 * @self: An instance of #APSkill
 * @bonus: (gboolean): The skill score bonus for the @self instance
 *
 * Sets the skill score bonus for @self.
 */
void
ap_skill_set_proficient (APSkill  *self,
                         gboolean  proficient)
{
    GValue new_proficient = G_VALUE_INIT;

    g_assert (AP_IS_SKILL (self));

    g_value_init (&new_proficient, G_TYPE_BOOLEAN);
    g_value_set_boolean (&new_proficient, proficient);

    g_object_set_property (G_OBJECT (self), "proficient", &new_proficient);

    g_value_unset (&new_proficient);
}

/**
 * ap_skill_get_proficient:
 * @self: An instance of #APSkill
 *
 * Returns if the character is proficient in the @self skill.
 *
 * Returns: (gboolean) (nullable): %TRUE or %FALSE if character is proficient for @self (#APSkill)
 */
gboolean
ap_skill_get_proficient (APSkill *self)
{
    GValue val = G_VALUE_INIT;
    gboolean retval;

    g_assert (AP_IS_SKILL (self));

    g_value_init (&val, G_TYPE_BOOLEAN);

    g_object_get_property (G_OBJECT (self), "proficient", &val);

    retval = g_value_get_boolean (&val);

    g_value_unset (&val);

    return retval;
}
